// Initial I2S_transmitter code, untested, by nockieboy August 2022.

module I2S_transmitter #(

    parameter int           BITS     = 16,  // Audio width, can be anything from 8 to 24
    parameter bit           INV_BCLK = 0    // When set, make I2S_BCLK fall during valid data

)(

    //Inputs
    input  logic            clk_in,         // High speed clock
    input  logic            clk_i2s,        // 50/50 duty cycle serial audio clock to feed through
    input  logic            clk_i2s_pulse,  // Should strobe for one clk_in cycle at the beginning of each new clk_i2s
    input  logic            sample_in,      // Optional input to reset the sample position.  This should either be tied to GND or only pulse once every 64 'clk_i2s_pulse's
    input  logic [BITS-1:0] DAC_Left,       // Left channel digital audio sampled once every 'sample_pulse' output
    input  logic [BITS-1:0] DAC_Right,      // Right channel digital audio sampled once every 'sample_pulse' output

    //Outputs
    output logic            sample_pulse,   // Pulses once when a new stereo sample is taken from the DAC_Left/Right inputs.  Hint: once every 64 clk_i2s_pulse's
    output logic            I2S_BCLK,       // I2S serial bit clock output (SCLK), basically the clk_i2s input in the correct phase
    output logic            I2S_WCLK,       // I2S !left / right output (LRCLK)
    output logic            I2S_DATA        // Serial data output

);

    localparam bit           LEFT     = 1'b0  ;
    localparam bit           RIGHT    = 1'b1  ;

    logic [$clog2(BITS)-1:0] bitcounter       ;
    logic              [5:0] I2S_counter <= 0 ;
    logic         [BITS-1:0] DAC_Buffer  <= 0 ;

    // Since I2S uses 64 clock pulses, we need one 6-bit reg counter.
    // I expect a serial shift out register plus a second register to buffer the right DAC input.
    // The code should be a few embedded if(), 1 for the main counter, one for the serial load/shifter,
    // and a last reg buffering the output regs.

    // We will copy/rename the PSG and make a new module testing the serializer with our fp_divider driving
    // the serializer to show our I2S output in the waveform view.  The left and right inputs should be
    // created in the _tb module.

    always_ff @( posedge clk_in ) begin

        // Invert clk_i2s for correct I2S_BCLK phase
        I2S_BCLK <= !clk_i2s ;

        // Manage I2S pulse counter and serialisation
        if ( clk_i2s_pulse ) begin

            bitcounter   <= bitcounter - 1           ;
            I2S_counter  <= I2S_counter + 1          ;
            I2S_DATA     <= DAC_Buffer[ bitcounter ] ;

        end

        // Strobe sample_pulse once every 64 cycles & sample appropriate L/R input
        if ( I2S_counter == 0 ) begin

            DAC_Buffer   <= DAC_Left  ;
            sample_pulse <= 1'b1      ;
            bitcounter   <= ~0        ; // set bitcounter to all 1's
            I2S_WCLK     <= LEFT      ;

        end else if ( I2S_counter == 32 ) begin

            DAC_Buffer   <= DAC_Right ;
            sample_pulse <= 1'b1      ;
            bitcounter   <= ~0        ; // set bitcounter to all 1's
            I2S_WCLK     <= RIGHT     ;
        
        end else begin

            sample_pulse <= 1'b0      ;

        end

    end

endmodule
